/*
  ============================================================================
   File: FUNCG722.H                                  v3.0 - 10/Jan/2007
  ============================================================================

      UGST/ITU-T G722 MODULE

          GLOBAL FUNCTION PROTOTYPES

     --------------------------------------
      (C) Copyright CNET Lannion A TSS/CMC
     --------------------------------------

   Original Author:
   
    J-P PETIT 
    CNET - Centre Lannion A
    LAA-TSS                         Tel: +33-96-05-39-41
    Route de Tregastel - BP 40      Fax: +33-96-05-13-16
    F-22301 Lannion CEDEX           Email: petitjp@lannion.cnet.fr
    FRANCE
    
   History:
   14.Mar.95    v1.0    Released for use ITU-T UGST software package Tool
                        based on the CNET's 07/01/90 version 2.00
   01.Jul.95    v2.0    Smart prototypes that work with many compilers; 
                        reformated; state variable structure added. 
24.Nov.15 - LPDSP32 optimizations by Ivo Coenen:
- (32-bit) int for Word16 type instead of (16-bit) short
- mult_long multiplications in qmf_tx and qmf_rx functions
- indexing instead of pointer incrementing inside loops to get dualld instructions
- using circular buffers for FIR delay lines
- static inline functions
- hardware saturation in qmf_tx and qmf_rx functions
- shifting inside limit function
- binary search in quantl function
- numerous minor optimizations
- input/output code is unsigned char type for easier handling
02.Sep.16 - LPDSP32 optimizations by Ivo Coenen:
- changed instruction order in upzero function
  ============================================================================
*/
#ifndef FUNCG722_H
#define FUNCG722_H 200

/* DEFINITION FOR SMART PROTOTYPES */
#ifndef ARGS
#if (defined(__STDC__) || defined(VMS) || defined(__DECC)  || defined(MSDOS) || defined(__MSDOS__)) || defined (__CYGWIN__) || defined (_MSC_VER)
#define ARGS(x) x
#else /* Unix: no parameters in prototype! */
#define ARGS(x) ()
#endif
#endif

#ifdef FLATTEN_LOOPS
#define CHESS_FLATTEN_LOOP chess_flatten_loop
#else
#define CHESS_FLATTEN_LOOP
#endif

/*___________________________________________________________________________
                                                                            
    Function Name : limit                                                   
                                                                            
    Purpose :                                                               
                                                                            
     .                  
                                                                            
    Inputs :                                                                
                                                                            
                                                                            
    Outputs :                                                               
                                                                            
     none                                                                   
                                                                            
    Return Value :                                                          
                                                                            
 ___________________________________________________________________________
*/
static inline Word16 limit (Word16 rl)
{
//  return (Word16)min(max(-16384,rl),16383);
  return (Word16)(rnd_saturate(to_accum(rl)<<17)>>17);
}
/* ..................... End of limit() ..................... */

/*___________________________________________________________________________
                                                                            
    Function Name : quantl                                                  
                                                                            
    Purpose :                                                               
                                                                            
     .                  
                                                                            
    Inputs :                                                                
                                                                            
                                                                            
    Outputs :                                                               
                                                                            
     none                                                                   
                                                                            
    Return Value :                                                          
                                                                            
 ___________________________________________________________________________
*/

static inline Word16 quantl (Word16 el, Word16 detl)
{
 /* table to read IL frorm SIL and MIL: misil(sil(0,1),mil(1,31)) */

  static char chess_storage(DMB) misil[64] = // Store in DMB to balance memory footprint
  {
    0x00, 0x3D, 0x3C, 0x3B, 0x3A, 0x39, 0x38, 0x37,
    0x36, 0x35, 0x34, 0x33, 0x32, 0x31, 0x30, 0x2F,
    0x2E, 0x2D, 0x2C, 0x2B, 0x2A, 0x29, 0x28, 0x27,
    0x26, 0x25, 0x24, 0x23, 0x22, 0x21, 0x20, 0x00,
    0x00, 0x3F, 0x3E, 0x1F, 0x1E, 0x1D, 0x1C, 0x1B,
    0x1A, 0x19, 0x18, 0x17, 0x16, 0x15, 0x14, 0x13,
    0x12, 0x11, 0x10, 0x0F, 0x0E, 0x0D, 0x0C, 0x0B,
    0x0A, 0x09, 0x08, 0x07, 0x06, 0x05, 0x04, 0x00
  };

   static Word32 chess_storage(DMB) q6shl3[30] = // Store in DMB to balance memory footprint
  {
    35*(1<<19), 72*(1<<19), 110*(1<<19), 150*(1<<19), 190*(1<<19), 233*(1<<19), 276*(1<<19),
    323*(1<<19), 370*(1<<19), 422*(1<<19), 473*(1<<19), 530*(1<<19), 587*(1<<19), 650*(1<<19), 714*(1<<19),
    786*(1<<19), 858*(1<<19), 940*(1<<19), 1023*(1<<19), 1121*(1<<19), 1219*(1<<19), 1339*(1<<19), 1458*(1<<19),
    1612*(1<<19), 1765*(1<<19), 1980*(1<<19), 2195*(1<<19), 2557*(1<<19), 2919*(1<<19), 3200*(1<<19)
  };

  Word32 sil, mil;
  accum_t wd, val;
  sil = ((unsigned long)el)>>31;
//  wd = (sil == 0) ? el : (MAX_16-(el & MAX_16));
  wd = update_low(to_accum(abs(el) - sil), -1);

  // Linear search
//  for(mil=0;mil<30;) {
//    val = fract_mult(q6shl3[mil++], (Word32)detl);
//    if(val>wd) {
//      break;
//    }
//  }

  // Binary search
  if (fract_mult(q6shl3[15-1], (Word32)detl)>wd) {
    // q6shl3[15]*detl>wd: mil is 15 or lower
    if (fract_mult(q6shl3[7-1], (Word32)detl)>wd) {
      // q6shl3[7]*detl>wd: mil is 7 or lower
      if (fract_mult(q6shl3[3-1], (Word32)detl)>wd) {
        // q6shl3[3]*detl>wd: mil is 3 or lower
        if (fract_mult(q6shl3[1-1], (Word32)detl)>wd) {
          // q6shl3[1]*detl>wd: mil is 1
          mil = 1;
        } else {
          // q6shl3[1]*detl<=wd: mil is 2 or higher
          if (fract_mult(q6shl3[2-1], (Word32)detl)>wd) {
            // q6shl3[2]*detl>wd: mil is 2
            mil = 2;
          } else {
            // q6shl3[2]*detl<=wd: mil is 3
            mil = 3;
          }
        }
      } else {
        // q6shl3[3]*detl<=wd: mil is 4 or higher
        if (fract_mult(q6shl3[5-1], (Word32)detl)>wd) {
          // q6shl3[5]*detl>wd: mil is 5 or lower
          if (fract_mult(q6shl3[4-1], (Word32)detl)>wd) {
            // q6shl3[4]*detl>wd: mil is 4
            mil = 4;
          } else {
            // q6shl3[4]*detl<=wd: mil is 5
            mil = 5;
          }
        } else {
          // q6shl3[5]*detl<=wd: mil is 6 or higher
          if (fract_mult(q6shl3[6-1], (Word32)detl)>wd) {
            // q6shl3[6]*detl>wd: mil is 6
            mil = 6;
          } else {
            // q6shl3[6]*detl<=wd: mil is 7
            mil = 7;
          }
        }
      }
    } else {
      // q6shl3[7]*detl<=wd: mil is 8 or higher
      if (fract_mult(q6shl3[11-1], (Word32)detl)>wd) {
        // q6shl3[11]*detl>wd: mil is 11 or lower
        if (fract_mult(q6shl3[9-1], (Word32)detl)>wd) {
          // q6shl3[9]*detl>wd: mil is 9 or lower
          if (fract_mult(q6shl3[8-1], (Word32)detl)>wd) {
            mil = 8;
            // q6shl3[8]*detl>wd: mil is 8
          } else {
            // q6shl3[8]*detl<=wd: mil is 9
            mil = 9;
          }
        } else {
          // q6shl3[9]*detl<=wd: mil is 10 or higher
          if (fract_mult(q6shl3[10-1], (Word32)detl)>wd) {
            // q6shl3[10]*detl>wd: mil is 10
            mil = 10;
          } else {
            // q6shl3[10]*detl<=wd: mil is 11
            mil = 11;
          }
        }
      } else {
        // q6shl3[11]*detl<=wd: mil is 12 or higher
        if (fract_mult(q6shl3[13-1], (Word32)detl)>wd) {
          // q6shl3[13]*detl>wd: mil is 13 or lower
          if (fract_mult(q6shl3[12-1], (Word32)detl)>wd) {
            // q6shl3[12]*detl>wd: mil is 12
            mil = 12;
          } else {
            // q6shl3[12]*detl<=wd: mil is 13
            mil = 13;
          }
        } else {
          // q6shl3[13]*detl<=wd: mil is 14 or higher
          if (fract_mult(q6shl3[14-1], (Word32)detl)>wd) {
            // q6shl3[14]*detl>wd: mil is 14
            mil = 14;
          } else {
            // q6shl3[14]*detl<=wd: mil is 15
            mil = 15;
          }
        }
      }
    }
  } else {
    // q6shl3[15]*detl<=wd: mil is 16 or higher
    if (fract_mult(q6shl3[23-1], (Word32)detl)>wd) {
      // q6shl3[23]*detl>wd: mil is 23 or lower
      if (fract_mult(q6shl3[19-1], (Word32)detl)>wd) {
        // q6shl3[19]*detl>wd: mil is 19 or lower
        if (fract_mult(q6shl3[17-1], (Word32)detl)>wd) {
          // q6shl3[17]*detl>wd: mil is 17 or lower
          if (fract_mult(q6shl3[16-1], (Word32)detl)>wd) {
            // q6shl3[16]*detl>wd: mil is 16
            mil = 16;
          } else {
            // q6shl3[16]*detl<=wd: mil is 17
            mil = 17;
          }
        } else {
          // q6shl3[17]*detl<=wd: mil is 18 or higher
          if (fract_mult(q6shl3[18-1], (Word32)detl)>wd) {
            // q6shl3[18]*detl>wd: mil is 18
            mil = 18;
          } else {
            // q6shl3[18]*detl<=wd: mil is 19
            mil = 19;
          }
        }
      } else {
        // q6shl3[19]*detl<=wd: mil is 20 or higher
        if (fract_mult(q6shl3[21-1], (Word32)detl)>wd) {
          // q6shl3[21]*detl>wd: mil is 21 or lower
          if (fract_mult(q6shl3[20-1], (Word32)detl)>wd) {
            // q6shl3[20]*detl>wd: mil is 20
            mil = 20;
          } else {
            // q6shl3[20]*detl<=wd: mil is 21
            mil = 21;
          }
        } else {
          // q6shl3[21]*detl<=wd: mil is 22 or higher
          if (fract_mult(q6shl3[22-1], (Word32)detl)>wd) {
            // q6shl3[22]*detl>wd: mil is 22
            mil = 22;
          } else {
            // q6shl3[22]*detl<=wd: mil is 23
            mil = 23;
          }
        }
      }
    } else {
      // q6shl3[23]*detl<=wd: mil is 24 or higher
      if (fract_mult(q6shl3[27-1], (Word32)detl)>wd) {
        // q6shl3[27]*detl>wd: mil is 27 or lower
        if (fract_mult(q6shl3[25-1], (Word32)detl)>wd) {
          // q6shl3[25]*detl>wd: mil is 25 or lower
          if (fract_mult(q6shl3[24-1], (Word32)detl)>wd) {
            // q6shl3[24]*detl>wd: mil is 24
            mil = 24;
          } else {
            // q6shl3[24]*detl<=wd: mil is 25
            mil = 25;
          }
        } else {
          // q6shl3[25]*detl<=wd: mil is 26 or higher
          if (fract_mult(q6shl3[26-1], (Word32)detl)>wd) {
            // q6shl3[26]*detl>wd: mil is 26
            mil = 26;
          } else {
            // q6shl3[26]*detl<=wd: mil is 27
            mil = 27;
          }
        }
      } else {
        // q6shl3[27]*detl<=wd: mil is 28 or higher
        if (fract_mult(q6shl3[29-1], (Word32)detl)>wd) {
          // q6shl3[29]*detl>wd: mil is 29 or lower
          if (fract_mult(q6shl3[28-1], (Word32)detl)>wd) {
            // q6shl3[28]*detl>wd: mil is 28
            mil = 28;
          } else {
            // q6shl3[28]*detl<=wd: mil is 29
            mil = 29;
          }
        } else {
          // q6shl3[29]*detl<=wd: mil is 30
          mil = 30;
        }
      }
    }
  }
  return (misil[(sil << 5) + mil]);
}
/* ..................... End of quantl() ..................... */


/*___________________________________________________________________________
                                                                            
    Function Name quanth:                                                   
                                                                            
    Purpose :                                                               
                                                                            
     .                  
                                                                            
    Inputs :                                                                
                                                                            
                                                                            
    Outputs :                                                               
                                                                            
     none                                                                   
                                                                            
    Return Value :                                                          
                                                                            
 ___________________________________________________________________________
*/
static inline Word16 quanth (Word16 eh, Word16 deth)
{
//  static Word16 misih[2][2] =
//  {
//    {1, 0},
//    {3, 2}
//  };
  static Word32 q2 = 564*(1<<19);

  Word32    sih, result;
  accum_t   wd, val;

  sih = ((unsigned long)eh)>>31;
  result = (3 >> sih);
//  wd = (sih == 0) ? eh : (MAX_16-(eh & MAX_16));  
  wd = update_low(to_accum(abs(eh) - sih), -1);
  val = fract_mult(q2, deth);
  if (val<wd) {
    return (--result);
  }
  return (result);
}
/* ..................... End of quanth() ..................... */


/*___________________________________________________________________________
                                                                            
    Function Name : invqal                                                  
                                                                            
    Purpose :                                                               
                                                                            
     .                  
                                                                            
    Inputs :                                                                
                                                                            
                                                                            
    Outputs :                                                               
                                                                            
     none                                                                   
                                                                            
    Return Value :                                                          
                                                                            
 ___________________________________________________________________________
*/
static inline Word16 invqal (Word16 il, Word16 detl)
{
//  /* inverse quantizer 4 bits for encoder or decoder */
//  static char   ril4[16] =
//  {0, 7, 6, 5, 4, 3, 2, 1, 7, 6, 5, 4, 3, 2, 1, 0};
//  static char   risil[16] =
//  {0, -1, -1, -1, -1, -1, -1, -1, 0, 0, 0, 0, 0, 0, 0, 0};
//  static short   oq4[8] =
//  {0, 150, 323, 530, 786, 1121, 1612, 2557};
//
//  /* Local variables */
//  Word16          ril, wd1, wd2;
//
//  ril = il >>2;
//  wd1 = (oq4[ril4[ril]])<<3;
//  wd2 = (risil[ril])? negate(wd1) : wd1;
//
//  return (mult(detl, wd2));

  static int   invqal_lookup[16] = {
    0*(1<<19), -2557*(1<<19), -1612*(1<<19), -1121*(1<<19), -786*(1<<19), -530*(1<<19), -323*(1<<19), -150*(1<<19), 2557*(1<<19), 1612*(1<<19), 1121*(1<<19), 786*(1<<19), 530*(1<<19), 323*(1<<19), 150*(1<<19), 0*(1<<19)};

//  accum_t acc = fract_mult(detl<<16, invqal_lookup[il>>2]);
//  return rnd_saturate(acc)>>16;
  return extract_high(fract_mult(detl, invqal_lookup[il>>2]));
}
/* ..................... End of invqal() ..................... */


/*___________________________________________________________________________
                                                                            
    Function Name : invqbl                                                  
                                                                            
    Purpose :                                                               
                                                                            
     .                  
                                                                            
    Inputs :                                                                
                                                                            
                                                                            
    Outputs :                                                               
                                                                            
     none                                                                   
                                                                            
    Return Value :                                                          
                                                                            
 ___________________________________________________________________________
*/
static inline Word16 invqbl (Word16 ilr, Word16 detl, Word16 mode)
{
//  /* Inverse quantizer 4, 5, and 6 bit tables for the decoder */
//  static char    ril4[16] = {0, 7, 6, 5, 4, 3, 2, 1, 7, 6, 5, 4, 3, 2, 1, 0};
//  static char    risi4[16] = {0, -1, -1, -1, -1, -1, -1, -1, 
//                               0, 0, 0, 0, 0, 0, 0, 0};
//  static short   oq4[8] = {0, 150, 323, 530, 786, 1121, 1612, 2557};
//  static char    ril5[32] = {
//    1, 1, 15, 14, 13, 12, 11, 10, 9, 8, 7, 6, 5, 4, 3, 2,
//    15, 14, 13, 12, 11, 10, 9, 8, 7, 6, 5, 4, 3, 2, 1, 1};
//
//  static char    risi5[32] = {
//    -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
//    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1};
//
//  static short   oq5[16] = {
//    0, 35, 110, 190, 276, 370, 473, 587,
//    714, 858, 1023, 1219, 1458, 1765, 2195, 2919};
//
//  static char    ril6[64] = {
//    1, 1, 1, 1, 30, 29, 28, 27, 26, 25, 24, 23, 22, 21, 20,
//    19, 18, 17, 16, 15, 14, 13, 12, 11, 10, 9, 8, 7, 6, 5, 4, 3,
//    30, 29, 28, 27, 26, 25, 24, 23, 22, 21, 20,
//    19, 18, 17, 16, 15, 14, 13, 12, 11, 10, 9, 8, 7, 6, 5, 4, 3, 2, 1, 2, 1};
//
//  static char    risi6[64] = {
//    -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
//    -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
//    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
//    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1, -1};
//
//  static short   oq6[31] = {
//    0, 17, 54, 91, 130, 170, 211, 254, 300, 347, 396, 447, 501,
//    558, 618, 682, 750, 822, 899, 982, 1072, 1170, 1279, 1399,
//    1535, 1689, 1873, 2088, 2376, 2738, 3101};
//
//  /* Local variables */
//  Word16         ril, wd1, wd2;
//
//  SWITCH (mode)
//  {
//  case 0:
//  case 1:
//      ril = ilr;
//      wd1 = (oq6[ril6[ril]])<< 3;
//      wd2 = (risi6[ril])?-wd1:wd1;
//      break;
//  case 2:
//      ril = ilr>> 1;
//      wd1 = (oq5[ril5[ril]])<< 3;
//      wd2 = (risi5[ril])?-wd1:wd1;
//    break;
// case 3:
//  default:
//    /* default to mode 3 */
//      ril = ilr>>2;
//      wd1 = (oq4[ril4[ril]])<< 3;
//      wd2 = (risi4[ril])?-wd1:wd1;
//    break;
//  }
//
//  return (mult(detl, wd2));

  int wd2;

  static short chess_storage(DMB) invqbl_lookup123[3*64] = { // Store in DMB to balance memory footprint
  // invqbl_lookup1
    -17*8, -17*8, -17*8, -17*8, -3101*8, -2738*8, -2376*8, -2088*8, -1873*8, -1689*8, -1535*8, -1399*8, -1279*8, -1170*8, -1072*8,
    -982*8, -899*8, -822*8, -750*8, -682*8, -618*8, -558*8, -501*8, -447*8, -396*8, -347*8, -300*8, -254*8, -211*8, -170*8, -130*8, -91*8,
    3101*8, 2738*8, 2376*8, 2088*8, 1873*8, 1689*8, 1535*8, 1399*8, 1279*8, 1170*8, 1072*8,
    982*8, 899*8, 822*8, 750*8, 682*8, 618*8, 558*8, 501*8, 447*8, 396*8, 347*8, 300*8, 254*8, 211*8, 170*8, 130*8, 91*8, 54*8, 17*8, -54*8, -17*8,
  // invqbl_lookup2 (each entry is doubled)
    -35*8, -35*8,
    -35*8, -35*8,
    -2919*8, -2919*8,
    -2195*8, -2195*8,
    -1765*8, -1765*8,
    -1458*8, -1458*8,
    -1219*8, -1219*8,
    -1023*8, -1023*8,
    -858*8, -858*8,
    -714*8, -714*8,
    -587*8, -587*8,
    -473*8, -473*8,
    -370*8, -370*8,
    -276*8, -276*8,
    -190*8, -190*8,
    -110*8, -110*8,
    2919*8, 2919*8,
    2195*8, 2195*8,
    1765*8, 1765*8,
    1458*8, 1458*8,
    1219*8, 1219*8,
    1023*8, 1023*8,
    858*8, 858*8,
    714*8, 714*8,
    587*8, 587*8,
    473*8, 473*8,
    370*8, 370*8,
    276*8, 276*8,
    190*8, 190*8,
    110*8, 110*8,
    35*8, 35*8,
    -35*8, -35*8,
  // invqbl_lookup3 (each entry is quadrupled)
    0*8, 0*8, 0*8, 0*8,
    -2557*8, -2557*8, -2557*8, -2557*8,
    -1612*8, -1612*8, -1612*8, -1612*8,
    -1121*8, -1121*8, -1121*8, -1121*8,
    -786*8, -786*8, -786*8, -786*8,
    -530*8, -530*8, -530*8, -530*8,
    -323*8, -323*8, -323*8, -323*8,
    -150*8, -150*8, -150*8, -150*8,
    2557*8, 2557*8, 2557*8, 2557*8,
    1612*8, 1612*8, 1612*8, 1612*8,
    1121*8, 1121*8, 1121*8, 1121*8,
    786*8, 786*8, 786*8, 786*8,
    530*8, 530*8, 530*8, 530*8,
    323*8, 323*8, 323*8, 323*8,
    150*8, 150*8, 150*8, 150*8,
    0*8, 0*8, 0*8, 0*8
  };

  wd2=invqbl_lookup123[ilr+(mode<<6)-64];

  //return (mult(detl, wd2));
  return extract_high(fract_mult(detl, wd2<<16)); // ICo: By making the lookup tables 32-bit the shift can be avoided, but this only wins 0.01 Mcps @ 16 kHz at the cost of 384 bytes of data memory
}
/* ..................... End of invqbl() ..................... */


/*___________________________________________________________________________
                                                                            
    Function Name : invqah                                                  
                                                                            
    Purpose :                                                               
                                                                            
     .                  
                                                                            
    Inputs :                                                                
                                                                            
                                                                            
    Outputs :                                                               
                                                                            
     none                                                                   
                                                                            
    Return Value :                                                          
                                                                            
 ___________________________________________________________________________
*/
static inline Word16 invqah (Word16 ih, Word16 deth)
{
//  static int     ih2[4] = {2, 1, 2, 1};
//  static int     sih[4] = {-1, -1, 0, 0};
//  static int     oq2[3] = {0, 202, 926};
//
//  Word16         wd1, wd2;
//
//  wd1 = (oq2[ih2[ih]])<< 3;
//  wd2 = (sih[ih])?-wd1:wd1;
//
//  return mult (wd2, deth);

  static int invqah_lookup[4] = {-926*(1<<19), -202*(1<<19), 926*(1<<19), 202*(1<<19)};

//  accum_t acc = fract_mult(deth<<16, invqah_lookup[ih]);
//  return rnd_saturate(acc)>>16;
  return extract_high(fract_mult(deth, invqah_lookup[ih]));
}
/* ..................... End of invqah() ..................... */


/*___________________________________________________________________________
                                                                            
    Function Name : logsch                                                  
                                                                            
    Purpose :                                                               
                                                                            
     .                  
                                                                            
    Inputs :                                                                
                                                                            
                                                                            
    Outputs :                                                               
                                                                            
     none                                                                   
                                                                            
    Return Value :                                                          
                                                                            
 ___________________________________________________________________________
*/
static inline Word16 logsch (Word16 ih, Word16 nbh)
{
//  static int     ih2[4] = {2, 1, 2, 1};
//  static int     wh[3]  = {0, -214, 798};
  static short logsch_lookup[4] = {798, -214, 798, -214};

  Word16         wd, nbph;
  
  //wd = mult (nbh, 32512);
  wd = extract_high(fract_mult(nbh, 32512<<16));
//  nbph = add (wd, wh[ih2[ih]]);
  nbph = add (wd, logsch_lookup[ih]);
  nbph = min(max(0,nbph),22528);
  
  return (nbph);
}
/* ..................... End of logsch() ..................... */


/*___________________________________________________________________________
                                                                            
    Function Name : logscl                                                  
                                                                            
    Purpose :                                                               
                                                                            
     .                  
                                                                            
    Inputs :                                                                
                                                                            
                                                                            
    Outputs :                                                               
                                                                            
     none                                                                   
                                                                            
    Return Value :                                                          
                                                                            
 ___________________________________________________________________________
*/
static inline Word16 logscl (Word16 il, Word16 nbl)
{
//  static char    ril4[16] = {0, 7, 6, 5, 4, 3, 2, 1, 7, 6, 5, 4, 3, 2, 1, 0};
//  static short   wl[8]    = {-60, -30, 58, 172, 334, 538, 1198, 3042};
  static short logscl_lookup[16] = {-60, 3042, 1198, 538, 334, 172, 58, -30, 3042, 1198, 538, 334, 172, 58, -30, -60};

  Word16         ril, wd, nbpl, il4;

//  ril = shr (il, 2);
//  wd = mult (nbl, 32512);
  wd = extract_high(fract_mult(nbl, 32512<<16));
//  il4 = ril4[ril];
//  nbpl = add (wd, wl[il4]);
  nbpl = add (wd, logscl_lookup[il>>2]);
  nbpl = min(max(0,nbpl),18432);

  return (nbpl);
}
/* ..................... End of logscl() ..................... */


/* ************** Table ILA used by scalel and scaleh ******************** */
static short chess_storage(DMB) ila[353] = // Store in DMB to balance memory footprint
  {
    (1+1)<<2, (1+1)<<2, (1+1)<<2, (1+1)<<2, (1+1)<<2, (1+1)<<2, (1+1)<<2, (1+1)<<2,
    (1+1)<<2, (1+1)<<2, (1+1)<<2, (1+1)<<2, (1+1)<<2, (1+1)<<2, (1+1)<<2, (1+1)<<2,
    (1+1)<<2, (1+1)<<2, (1+1)<<2, (2+1)<<2, (2+1)<<2, (2+1)<<2, (2+1)<<2, (2+1)<<2,
    (2+1)<<2, (2+1)<<2, (2+1)<<2, (2+1)<<2, (2+1)<<2, (2+1)<<2, (2+1)<<2, (2+1)<<2,
    (3+1)<<2, (3+1)<<2, (3+1)<<2, (3+1)<<2, (3+1)<<2, (3+1)<<2, (3+1)<<2, (3+1)<<2,
    (3+1)<<2, (3+1)<<2, (3+1)<<2, (4+1)<<2, (4+1)<<2, (4+1)<<2, (4+1)<<2, (4+1)<<2,
    (4+1)<<2, (4+1)<<2, (4+1)<<2, (5+1)<<2, (5+1)<<2, (5+1)<<2, (5+1)<<2, (5+1)<<2,
    (5+1)<<2, (5+1)<<2, (6+1)<<2, (6+1)<<2, (6+1)<<2, (6+1)<<2, (6+1)<<2, (6+1)<<2,
    (7+1)<<2, (7+1)<<2, (7+1)<<2, (7+1)<<2, (7+1)<<2, (7+1)<<2, (8+1)<<2, (8+1)<<2,
    (8+1)<<2, (8+1)<<2, (8+1)<<2, (9+1)<<2, (9+1)<<2, (9+1)<<2, (9+1)<<2, (10+1)<<2,
    (10+1)<<2, (10+1)<<2, (10+1)<<2, (11+1)<<2, (11+1)<<2, (11+1)<<2, (11+1)<<2, (12+1)<<2,
    (12+1)<<2, (12+1)<<2, (13+1)<<2, (13+1)<<2, (13+1)<<2, (13+1)<<2, (14+1)<<2, (14+1)<<2,
    (15+1)<<2, (15+1)<<2, (15+1)<<2, (16+1)<<2, (16+1)<<2, (16+1)<<2, (17+1)<<2, (17+1)<<2,
    (18+1)<<2, (18+1)<<2, (18+1)<<2, (19+1)<<2, (19+1)<<2, (20+1)<<2, (20+1)<<2, (21+1)<<2,
    (21+1)<<2, (22+1)<<2, (22+1)<<2, (23+1)<<2, (23+1)<<2, (24+1)<<2, (24+1)<<2, (25+1)<<2,
    (25+1)<<2, (26+1)<<2, (27+1)<<2, (27+1)<<2, (28+1)<<2, (28+1)<<2, (29+1)<<2, (30+1)<<2,
    (31+1)<<2, (31+1)<<2, (32+1)<<2, (33+1)<<2, (33+1)<<2, (34+1)<<2, (35+1)<<2, (36+1)<<2,
    (37+1)<<2, (37+1)<<2, (38+1)<<2, (39+1)<<2, (40+1)<<2, (41+1)<<2, (42+1)<<2, (43+1)<<2,
    (44+1)<<2, (45+1)<<2, (46+1)<<2, (47+1)<<2, (48+1)<<2, (49+1)<<2, (50+1)<<2, (51+1)<<2,
    (52+1)<<2, (54+1)<<2, (55+1)<<2, (56+1)<<2, (57+1)<<2, (58+1)<<2, (60+1)<<2, (61+1)<<2,
    (63+1)<<2, (64+1)<<2, (65+1)<<2, (67+1)<<2, (68+1)<<2, (70+1)<<2, (71+1)<<2, (73+1)<<2,
    (75+1)<<2, (76+1)<<2, (78+1)<<2, (80+1)<<2, (82+1)<<2, (83+1)<<2, (85+1)<<2, (87+1)<<2,
    (89+1)<<2, (91+1)<<2, (93+1)<<2, (95+1)<<2, (97+1)<<2, (99+1)<<2, (102+1)<<2, (104+1)<<2,
    (106+1)<<2, (109+1)<<2, (111+1)<<2, (113+1)<<2, (116+1)<<2, (118+1)<<2, (121+1)<<2, (124+1)<<2,
    (127+1)<<2, (129+1)<<2, (132+1)<<2, (135+1)<<2, (138+1)<<2, (141+1)<<2, (144+1)<<2, (147+1)<<2,
    (151+1)<<2, (154+1)<<2, (157+1)<<2, (161+1)<<2, (165+1)<<2, (168+1)<<2, (172+1)<<2, (176+1)<<2,
    (180+1)<<2, (184+1)<<2, (188+1)<<2, (192+1)<<2, (196+1)<<2, (200+1)<<2, (205+1)<<2, (209+1)<<2,
    (214+1)<<2, (219+1)<<2, (223+1)<<2, (228+1)<<2, (233+1)<<2, (238+1)<<2, (244+1)<<2, (249+1)<<2,
    (255+1)<<2, (260+1)<<2, (266+1)<<2, (272+1)<<2, (278+1)<<2, (284+1)<<2, (290+1)<<2, (296+1)<<2,
    (303+1)<<2, (310+1)<<2, (316+1)<<2, (323+1)<<2, (331+1)<<2, (338+1)<<2, (345+1)<<2, (353+1)<<2,
    (361+1)<<2, (369+1)<<2, (377+1)<<2, (385+1)<<2, (393+1)<<2, (402+1)<<2, (411+1)<<2, (420+1)<<2,
    (429+1)<<2, (439+1)<<2, (448+1)<<2, (458+1)<<2, (468+1)<<2, (478+1)<<2, (489+1)<<2, (500+1)<<2,
    (511+1)<<2, (522+1)<<2, (533+1)<<2, (545+1)<<2, (557+1)<<2, (569+1)<<2, (582+1)<<2, (594+1)<<2,
    (607+1)<<2, (621+1)<<2, (634+1)<<2, (648+1)<<2, (663+1)<<2, (677+1)<<2, (692+1)<<2, (707+1)<<2,
    (723+1)<<2, (739+1)<<2, (755+1)<<2, (771+1)<<2, (788+1)<<2, (806+1)<<2, (823+1)<<2, (841+1)<<2,
    (860+1)<<2, (879+1)<<2, (898+1)<<2, (918+1)<<2, (938+1)<<2, (958+1)<<2, (979+1)<<2, (1001+1)<<2,
    (1023+1)<<2, (1045+1)<<2, (1068+1)<<2, (1092+1)<<2, (1115+1)<<2, (1140+1)<<2, (1165+1)<<2, (1190+1)<<2,
    (1216+1)<<2, (1243+1)<<2, (1270+1)<<2, (1298+1)<<2, (1327+1)<<2, (1356+1)<<2, (1386+1)<<2, (1416+1)<<2,
    (1447+1)<<2, (1479+1)<<2, (1511+1)<<2, (1544+1)<<2, (1578+1)<<2, (1613+1)<<2, (1648+1)<<2, (1684+1)<<2,
    (1721+1)<<2, (1759+1)<<2, (1797+1)<<2, (1837+1)<<2, (1877+1)<<2, (1918+1)<<2, (1960+1)<<2, (2003+1)<<2,
    (2047+1)<<2, (2092+1)<<2, (2138+1)<<2, (2185+1)<<2, (2232+1)<<2, (2281+1)<<2, (2331+1)<<2, (2382+1)<<2,
    (2434+1)<<2, (2488+1)<<2, (2542+1)<<2, (2598+1)<<2, (2655+1)<<2, (2713+1)<<2, (2773+1)<<2, (2833+1)<<2,
    (2895+1)<<2, (2959+1)<<2, (3024+1)<<2, (3090+1)<<2, (3157+1)<<2, (3227+1)<<2, (3297+1)<<2, (3370+1)<<2,
    (3443+1)<<2, (3519+1)<<2, (3596+1)<<2, (3675+1)<<2, (3755+1)<<2, (3837+1)<<2, (3921+1)<<2, (4007+1)<<2,
    (4095+1)<<2};
/* ************************* End of Table ILA **************************** */


/*___________________________________________________________________________
                                                                            
    Function Name : scalel                                                  
                                                                            
    Purpose :                                                               
                                                                            
     .                  
                                                                            
    Inputs :                                                                
                                                                            
                                                                            
    Outputs :                                                               
                                                                            
     none                                                                   
                                                                            
    Return Value :                                                          
                                                                            
 ___________________________________________________________________________
*/
static inline Word16 scalel (Word16 nbpl)
{
  Word16    wd1, wd2;
  wd1 = (nbpl>>6) & 511;
  wd2 = wd1 + 64;
//  return ((ila[wd2] + 1) << 2);//stl2000
  return ila[wd2];
}
/* ..................... End of scalel() ..................... */

/*___________________________________________________________________________
                                                                            
    Function Name : scaleh                                                  
                                                                            
    Purpose :                                                               
                                                                            
     .                  
                                                                            
    Inputs :                                                                
                                                                            
                                                                            
    Outputs :                                                               
                                                                            
     none                                                                   
                                                                            
    Return Value :                                                          
                                                                            
 ___________________________________________________________________________
*/
static inline Word16 scaleh (Word16 nbph)
{
  Word16    wd;

  wd = (nbph>>6) & 511;
//  return ((ila[wd] + 1) << 2); //stl2000
  return ila[wd];
}
/* ..................... End of scaleh() ..................... */

static inline void uppol1 (Word16 al [], Word16 plt [])
{
  Word16    sg0, sg1, wd1, wd2, wd3, apl1;

  sg0 = plt[0]>>15;
  sg1 = plt[1]>>15;

  wd1 = (sg0 == sg1)?192:-192;
 
//  wd2 = mult (al[1], 32640);
  wd2 = extract_high(fract_mult(al[1], 32640<<16));
  apl1 = add (wd1, wd2);
  wd3 = sub (15360, al[2]); // Result cannot be -32768 (so following negate operation doesn't need saturation)

/*  apl1 = (apl1 > wd3) ? wd3 : ((apl1 < -wd3) ? negate (wd3) : apl1); */
  apl1 = min(max(-wd3,apl1),wd3);

  /* Shift of the plt signals */
  plt[2] = plt[1];
  plt[1] = plt[0];
  al[1] = apl1;
}
/* ..................... End of uppol1() ..................... */

/*___________________________________________________________________________
                                                                            
    Function Name : uppol2                                                  
                                                                            
    Purpose :                                                               
                                                                            
     .                  
                                                                            
    Inputs :                                                                
                                                                            
                                                                            
    Outputs :                                                               
                                                                            
     none                                                                   
                                                                            
    Return Value :                                                          
     None.                                                                  
 ___________________________________________________________________________
*/

static inline void uppol2 (Word16 al [], Word16 plt [])
{
  Word16    sg0, sg1, sg2, wd1, wd2, wd3, apl2;

  sg0 = plt[0]>> 15;
  sg1 = plt[1]>> 15;
  sg2 = plt[2]>> 15;

  wd1 = al[1];

  wd2 = (sg0 == sg1)? -wd1:wd1;

  wd2 = rnd_saturate(to_accum(wd2)<<(16+2))>>(16+7);

  wd3 = (sg0 == sg2) ? 128:-128;


//  apl2 = wd2 + wd3 + mult (al[2], 32512);
  apl2 = wd2 + wd3 + extract_high(fract_mult(al[2], 32512<<16));
  al[2] = min(max(-12288,apl2),12288);
}

/* ..................... End of uppol2() ..................... */


/*___________________________________________________________________________
                                                                            
    Function Name : upzero                                                  
                                                                            
    Purpose :                                                               
                                                                            
     .                  
                                                                            
    Inputs :                                                                
                                                                            
                                                                            
    Outputs :                                                               
                                                                            
     none                                                                   
                                                                            
    Return Value :                                                          
     None.                                                                  
                                                                            
 ___________________________________________________________________________
*/
const Word32 const_upzero = 0x7f800000;

static inline void upzero (Word16 dlt [], Word16 bl [])
{
  Word16    sg0, sgia, sgib, wd1, wd2a, wd2b, wd3a, wd3b;
  Word32    i;

  accum_t   a_prod;
  accum_t   b_prod;

  Word32    var1, var2;
  Word32    resa, resb;

  /* shift of the dlt line signal and update of bl */

  wd1 = (dlt[0] ) ? 128 : 0;

  sg0 = dlt[0]>>15;

  // Initialize var1 and var2
  var1 = bl[6];
  var2 = bl[5];

  for (i = 6; i > 0; i=i-2) chess_loop_count(3) CHESS_FLATTEN_LOOP
  {
    // This lines are interleaved to produce faster code
    a_prod = fract_mult(var1, const_upzero); 
    wd3a = rnd_saturate(a_prod)>>16;
    b_prod = fract_mult(var2, const_upzero);
    wd3b = rnd_saturate(b_prod)>>16;

    sgia = dlt[i]>>15;
    sgib = dlt[i-1]>>15;

    // This lines are interleaved to produce faster code
    dlt[i] = dlt[i-1];
    wd2a = (sg0 == sgia) ? wd1 : (-wd1);
    dlt[i-1] = dlt[i-2];
    wd2b = (sg0 == sgib) ? wd1 : (-wd1);

    // 16-bit result cannot saturate, because:
    // 0x7f80*0x7fff=0x7f7f + 128 = 0x7fff
    // 0x7f80*0x8000=0x8080 - 128 = 0x8000
    resa = wd2a + wd3a;
    resb = wd2b + wd3b;

    // Prepare next read before writing back to memory to avoid having a nop due to read after write hazard
    // This comes at the cost of two useless reads are performed during the last iternation of the loop
    var1 = bl[i-2];
    var2 = bl[i-3];

    // Store bl in a the upper part of the 32-bit word
    bl[i] = resa<<16;
    bl[i-1] = resb<<16;
  }
}
/* ..................... End of upzero() ..................... */


/*___________________________________________________________________________
                                                                            
    Function Name : filtep                                                  
                                                                            
    Purpose :                                                               
                                                                            
     .                  
                                                                            
    Inputs :                                                                
                                                                            
                                                                            
    Outputs :                                                               
                                                                            
     none                                                                   
                                                                            
    Return Value :                                                          
                                                                            
 ___________________________________________________________________________
*/
static inline Word16 filtep (Word16 rlt [], Word16 al [])
{

  Word16    wd1, wd2, spl;

  /* shift of rlt */
  rlt[2] = rlt[1];    
  rlt[1] = rlt[0];    

  wd1 = add (rlt[1], rlt[1]);
//  wd1 = mult (al[1], wd1);
  wd1 = extract_high(fract_mult(al[1], wd1<<16));
  wd2 = add (rlt[2], rlt[2]);
//  wd2 = mult (al[2], wd2);
  wd2 = extract_high(fract_mult(al[2], wd2<<16));
  spl = add (wd1, wd2);

  return (spl);
}



/* ..................... End of filtep() ..................... */


/*___________________________________________________________________________
                                                                            
    Function Name : filtez                                                  
                                                                            
    Purpose :                                                               
                                                                            
     .                  
                                                                            
    Inputs :                                                                
                                                                            
                                                                            
    Outputs :                                                               
                                                                            
     none                                                                   
                                                                            
    Return Value :                                                          
                                                                            
 ___________________________________________________________________________
*/
static inline Word16 filtez (Word16 dlt [], Word16 bl [])
{

  Word16    szl;
  accum_t   resa;
  accum_t   resb;

  szl = 0;

  for (int i = 6; i > 0; i=i-2) chess_loop_count(3) CHESS_FLATTEN_LOOP
  {
    resa = fract_mult(dlt[i]<<17, bl[i]); // dlt is 15-bit
    szl += rnd_saturate(resa)>>16;
    resb = fract_mult(dlt[i-1]<<17, bl[i-1]); // dlt is 15-bit
    szl += rnd_saturate(resb)>>16;
  }

  return szl;
}
/* ..................... End of filtez() ..................... */


/*___________________________________________________________________________
                                                                            
    Function Name : lsbcodecrs                                                  
                                                                            
    Purpose :                                                               
                                                                            
     Encoder/Decoder lower subband reset.                         
                                                                            
    Inputs :  
      s  - pointer to state variable (read/write)                           
                                                                            
    Return Value :                                                          
     none                                        
 ___________________________________________________________________________
*/
#define AL   s->al
#define BL   s->bl
#define DETL s->detl
#define DLT  s->dlt
#define NBL  s->nbl
#define PLT  s->plt
#define RLT  s->rlt
#define SL   s->sl
#define SPL  s->spl
#define SZL  s->szl

static inline Word16 lsbcodecrs (g722_state *s)
{
    DETL = 32;
    SL = SPL = SZL = NBL = 0;
    AL[1] = AL[2] = 0;
    BL[1] = BL[2] = BL[3] = BL[4] = BL[5] = BL[6] = 0;
    DLT[0] = DLT[1] = DLT[2] = DLT[3] = DLT[4] = DLT[5] = DLT[6] = 0;
    PLT[0] = PLT[1] = PLT[2] = 0;
    RLT[0] = RLT[1] = RLT[2] = 0;
    return (0);
}
#undef AL
#undef BL
#undef DETL
#undef DLT
#undef NBL
#undef PLT
#undef RLT
#undef SL
#undef SPL
#undef SZL
/* ........................ End of lsbcodrs() ........................ */

/*___________________________________________________________________________
                                                                            
    Function Name : lsbcod                                                  
                                                                            
    Purpose :                                                               
                                                                            
     Encode lower subband of incomung speech/music.                         
                                                                            
    Inputs :                                                                
      xl - Low sub-band portion of the input sample    
      s  - pointer to state variable (read/write)                           
                                                                            
    Return Value :                                                          
     Encoded sample as a 16-bit word                                        
 ___________________________________________________________________________
*/
#define AL   s->al
#define BL   s->bl
#define DETL s->detl
#define DLT  s->dlt
#define NBL  s->nbl
#define PLT  s->plt
#define RLT  s->rlt
#define SL   s->sl
#define SPL  s->spl
#define SZL  s->szl

static inline Word16 lsbcod (Word16 xl, g722_state *s)
{
  Word16          el, nbpl, il;
  el = sub (xl, SL);           /* subtra */
  il = quantl (el, DETL);
  DLT[0] = invqal (il, DETL); // Result is 15-bit
  nbpl = logscl (il, NBL);
  NBL = nbpl;
  DETL = scalel (nbpl); // Result <= 0x4000
  PLT[0] = add (DLT[0], SZL);   /* parrec */
  RLT[0] = add (SL, DLT[0]);    /* recons */
  upzero (DLT, BL);
  uppol2 (AL, PLT);
  uppol1 (AL, PLT);
  SZL = filtez (DLT, BL); // Assumes 15-bit input
  SPL = filtep (RLT, AL);
  SL = add (SPL, SZL);          /* predic */
  /* Return encoded sample */
  return (il);
}
#undef AL
#undef BL
#undef DETL
#undef DLT
#undef NBL
#undef PLT
#undef RLT
#undef SL
#undef SPL
#undef SZL
/* ........................ End of lsbcod() ........................ */






/*___________________________________________________________________________
                                                                            
    Function Name : hsbcodecrs                                                  
                                                                            
    Purpose :                                                               
                                                                            
     Encoder/Decoder upper sub-band reset.                     
                                                                            
    Inputs :            
      s  - pointer to state variable (read/write)                           
                                                                            
    Return Value :                                                          
  none                                              
 ___________________________________________________________________________
*/
#define AH   s->ah
#define BH   s->bh
#define DETH s->deth
#define DH   s->dh
#define NBH  s->nbh
#define PH   s->ph
#define RH   s->rh
#define SH   s->sh
#define SZH  s->szh

static inline Word16 hsbcodecrs (g722_state *s)
{
    DETH = 8;
    SH = SZH = NBH = 0;
    AH[1] = AH[2] = 0;
    BH[1] = BH[2] = BH[3] = BH[4] = BH[5] = BH[6] = 0;
    DH[0] = DH[1] = DH[2] = DH[3] = DH[4] = DH[5] = DH[6] = 0;
    PH[0] = PH[1] = PH[2] = 0;
    RH[0] = RH[1] = RH[2] = 0;
    return (0);
}
#undef AH
#undef BH
#undef DETH
#undef DH
#undef NBH
#undef PH
#undef RH
#undef SH
#undef SZH
/* ........................ End of hsbcodrs() ........................ */


/*___________________________________________________________________________
                                                                            
    Function Name : hsbcod                                                  
                                                                            
    Purpose :                                                               
                                                                            
     Encode upper sub-band of incoming speech or music.                     
                                                                            
    Inputs :                                                                
      xh - High sub-band portion of the input sample        
      s  - pointer to state variable (read/write)                           
                                                                            
    Return Value :                                                          
     ADPCM code for the high-band portion of the input sample as a          
     16-bit word.                                                           
                                                                            
 ___________________________________________________________________________
*/
#define AH   s->ah
#define BH   s->bh
#define DETH s->deth
#define DH   s->dh
#define NBH  s->nbh
#define PH   s->ph
#define RH   s->rh
#define SH   s->sh
#define SZH  s->szh

static inline Word16 hsbcod (Word16 xh, g722_state *s)
{
  Word16          eh, nbph, ih, sph;
  eh = sub (xh, SH);         /* subtra */
  ih = quanth (eh, DETH);
  DH[0] = invqah (ih, DETH); // Result is 15-bit
  nbph = logsch (ih, NBH);
  NBH = nbph;
  DETH = scaleh (nbph); // Result <= 0x4000
  PH[0] = add (DH[0], SZH);   /* parrec */
  RH[0] = add (SH, DH[0]);    /* recons */
  upzero (DH, BH);
  uppol2 (AH, PH);
  uppol1 (AH, PH);
  SZH = filtez (DH, BH); // Assumes 15-bit input
  sph = filtep (RH, AH);
  SH = add (sph, SZH);        /* predic */
  return (ih);
}
#undef AH
#undef BH
#undef DETH
#undef DH
#undef NBH
#undef PH
#undef RH
#undef SH
#undef SZH
/* ........................ End of hsbcod() ........................ */




/*___________________________________________________________________________
                                                                            
    Function Name : lsbdec                                                  
                                                                            
    Purpose :                                                               
                                                                            
     Decode lower subband of incomung speech/music.                         
                                                                            
    Inputs :                                                                
      ilr - ADPCM encoding of the low sub-band                              
      mode - G.722 operation mode     
      s   - pointer to state variable (read/write)                          
                                                                            
    Return Value :                                                          
     Decoded low-band portion of the recovered sample as a 16-bit word      
 ___________________________________________________________________________
*/
#define AL   s->al
#define BL   s->bl
#define DETL s->detl
#define DLT  s->dlt
#define NBL  s->nbl
#define PLT  s->plt
#define RLT  s->rlt
#define SL   s->sl
#define SPL  s->spl
#define SZL  s->szl

static inline Word16 lsbdec (Word16 ilr, Word16 mode, g722_state *s)
{
  Word16          dl, rl, nbpl, yl;
  dl = invqbl (ilr, DETL, mode);
  rl = add (SL, dl);              /* recons */
  yl = limit (rl);
  DLT[0] = invqal (ilr, DETL); // Result is 15-bit
  nbpl = logscl (ilr, NBL);
  NBL = nbpl;
  DETL = scalel (nbpl); // Result <= 0x4000
  PLT[0] = add (DLT[0], SZL);     /* parrec */
  RLT[0] = add (SL, DLT[0]);      /* recons */
  upzero (DLT, BL);
  uppol2 (AL, PLT);
  uppol1 (AL, PLT);
  SZL = filtez (DLT, BL); // Assumes 15-bit input
  SPL = filtep (RLT, AL);
  SL = add (SPL, SZL);            /* predic */
  return (yl);
}
#undef AL
#undef BL
#undef DETL
#undef DLT
#undef NBL
#undef PLT
#undef RLT
#undef SL
#undef SPL
#undef SZL
/* ........................ End of lsbdec() ........................ */


/*___________________________________________________________________________
                                                                            
    Function Name : hsbdec                                                  
                                                                            
    Purpose :                                                               
                                                                            
     Decode lower subband of incomung speech/music.                         
                                                                            
    Inputs :                                                                
      ih  - ADPCM encoding of the high sub-band  
      s   - pointer to state variable (read/write)                          
                                                                            
    Return Value :                                                          
     Decoded high-band portion of the recovered sample as a 16-bit word     
 ___________________________________________________________________________
*/
#define AH   s->ah
#define BH   s->bh
#define DETH s->deth
#define DH   s->dh
#define NBH  s->nbh
#define PH   s->ph
#define RH   s->rh
#define SH   s->sh
#define SZH  s->szh

static inline Word16 hsbdec (Word16 ih, g722_state *s)
{
  Word16          nbph, yh, sph;
  DH[0] = invqah (ih, DETH);
  nbph = logsch (ih, NBH);
  NBH = nbph;
  DETH = scaleh (nbph); // Result <= 0x4000
  PH[0] = add (DH[0], SZH);    /* add */
  RH[0] = add (SH, DH[0]);     /* recons */
  upzero (DH, BH);
  uppol2 (AH, PH);
  uppol1 (AH, PH);
  SZH = filtez (DH, BH);
  sph = filtep (RH, AH);
  SH = add (sph, SZH);         /* predic */
  yh = limit (RH[0]);
  return (yh);
}
#undef AH
#undef BH
#undef DETH
#undef DH
#undef NBH
#undef PH
#undef RH
#undef SH
#undef SZH
/* ........................ End of hsbdec() ........................ */


/* **** Coefficients for both transmission and reception QMF **** */

static Word32 chess_storage(DMB) coef_qmf[24] =

//  {
//    3 * 2, -11 * 2, -11 * 2, 53 * 2, 12 * 2, -156 * 2,
//    32 * 2, 362 * 2, -210 * 2, -805 * 2, 951 * 2, 3876 * 2,
//    3876 * 2, 951 * 2, -805 * 2, -210 * 2, 362 * 2, 32 * 2,
//    -156 * 2, 12 * 2, 53 * 2, -11 * 2, -11 * 2, 3 * 2};
// ICo: Coefficients are used in fractional multiplications and therefore do not need to be multiplied by two
//      Normalizing coefficients to lower the power consumption of the multipliers by keeping 19 LSBs at zero)
  {
    3*(1<<19), -11*(1<<19), -11*(1<<19), 53*(1<<19), 12*(1<<19), -156*(1<<19),
    32*(1<<19), 362*(1<<19), -210*(1<<19), -805*(1<<19), 951*(1<<19), 3876*(1<<19),
    3876*(1<<19), 951*(1<<19), -805*(1<<19), -210*(1<<19), 362*(1<<19), 32*(1<<19),
    -156*(1<<19), 12*(1<<19), 53*(1<<19), -11*(1<<19), -11*(1<<19), 3*(1<<19)};
/* ..................... End of table coef_qmf[] ..................... */


/*___________________________________________________________________________
                                                                            
    Function Name : qmf_tx                                                  
                                                                            
    Purpose :                                                               
                                                                            
     G722 QMF analysis (encoder) filter. Uses coefficients in array         
     coef_qmf[] defined above.                                              
                                                                            
    Inputs :                                                                
     xin0 - first sample for the QMF filter (read-only)                     
     xin1 - secon sample for the QMF filter (read-only)                     
     xl   - lower band portion of samples xin0 and xin1 (write-only)        
     xh   - higher band portion of samples xin0 and xin1 (write-only)       
     s    - pointer to state variable structure (read/write)                
                                                                            
    Return Value :                                                          
     None.                                                                  
 ___________________________________________________________________________
*/

#define delayx s->qmf_tx_delayx
#define pdelayx s->p_tx_delayx
static inline void qmf_tx (Word16 xin0, Word16 xin1, Word16 *xl, Word16 *xh, g722_state *s)
{

  /* Local variables */
  Word16          i,j,k;
  accum_t         accuma, accumb;
  Word32          comp_low, comp_high;

  /* Saving past samples in delay line */
  pdelayx = cyclic_add(pdelayx, -1, delayx, 24);
  *pdelayx = xin1;
  pdelayx = cyclic_add(pdelayx, -1, delayx, 24);
  *pdelayx = xin0;

  /* QMF filtering */
  accuma = to_accum(0);
  accumb = to_accum(0);

  for (i = 0; i < 24; i+=2) chess_loop_count(12) CHESS_FLATTEN_LOOP
  {
    accuma += fract_mult(coef_qmf[i]  , *pdelayx);
    pdelayx = cyclic_add(pdelayx, 1, delayx, 24);
    accumb += fract_mult(coef_qmf[i+1], *pdelayx);
    pdelayx = cyclic_add(pdelayx, 1, delayx, 24);
  }
  comp_low  = rnd_saturate((accuma + accumb)<<15)>>17;
  comp_high = rnd_saturate((accuma - accumb)<<15)>>17;

  *xl = extract_l(comp_low);
  *xh = extract_l(comp_high);
}
#undef delayx
#undef pdelayx
/* ..................... End of qmf_tx() ..................... */


/*___________________________________________________________________________
                                                                            
    Function Name : qmf_rx                                                  
                                                                            
     G722 QMF synthesis (decoder) filter. Uses coefficients in array        
     coef_qmf[] defined above.                                              
                                                                            
    Inputs :                                                                
     xout0 - first sample out of the QMF filter (write-only)                
     xout1 - secon sample out of the QMF filter (write-only)                
     rl    - lower band portion of a sample (read-only)                     
     rh    - higher band portion of a sample (read-only)                    
     s     - pointer to state variable structure (read/write)               
                                                                            
    Return Value :                                                          
     None.                                                                  
 ___________________________________________________________________________
*/

#define delayx s->qmf_rx_delayx
#define pdelayx s->p_rx_delayx
static inline void qmf_rx (Word16 rl, Word16 rh, Word16 *xout1, Word16 *xout2, g722_state *s)
{
  Word32          i;
  accum_t         accuma, accumb;
  Word32          comp_low, comp_high;

  /* compute sum and difference from lower-band (rl) and higher-band (rh) signals */
  /* update delay line */
  pdelayx = cyclic_add(pdelayx, -1, delayx, 24);
  *pdelayx = add (rl, rh);
  pdelayx = cyclic_add(pdelayx, -1, delayx, 24);
  *pdelayx = sub (rl, rh);

  /* qmf_rx filtering */
  accuma = to_accum(0);
  accumb = to_accum(0);

  for (i = 0; i < 24; i+=2) chess_loop_count(12) CHESS_FLATTEN_LOOP
  {
    accuma += fract_mult(coef_qmf[i]  , *pdelayx);
    pdelayx = cyclic_add(pdelayx, 1, delayx, 24);
    accumb += fract_mult(coef_qmf[i+1], *pdelayx);
    pdelayx = cyclic_add(pdelayx, 1, delayx, 24);
  }

  comp_low  = rnd_saturate(accuma<<17);
  comp_high = rnd_saturate(accumb<<17);

  /* compute output samples */
  *xout1 = extract_h(comp_low);
  *xout2 = extract_h(comp_high);
}

#undef delayx
#undef pdelayx


/* ..................... End of qmf_rx() ..................... */

#endif /* FUNCG722_H */
/* ........................ End of file funcg722.h ......................... */
