# ----------------------------------------------------------------------------
# Copyright (c) 2017 Semiconductor Components Industries, LLC (d/b/a
# ON Semiconductor), All Rights Reserved
#
# This code is the property of ON Semiconductor and may not be redistributed
# in any form without prior written permission from ON Semiconductor.
# The terms of use and warranty for this code are covered by contractual
# agreements between ON Semiconductor and the licensee.
#
# This is Reusable Code.
#
# -----------------------------------------------------------------------------
#

from os import path

from elftools.elf.elffile import ELFFile

from DSPMemoryIntegrator import DSPMemoryIntegrator
from DSPSymbolIntegrator import DSPSymbolIntegrator

class DSPIntegrator(object):
    '''
    This class will take a given ELF file and extract the memory and symbol
    information into C header and source files. It does this by using two
    subordinate objects to do the real work, this is primarily arranging the
    parameters for these objects.  
    '''

    def __init__(self, elf, output, tag, symbols, v1):
        '''
        Constructor for the class.
        
        elf
            Specifies an input ELF file which has been generated using the
            Synopsys LPDSP32 compiler tools.
        output
            Specifies a folder into which the generated files will be placed
        tag
            Defines a tag which will be added to each generated file and data
            structure to aid in identification in the C parent code.
        symbols 
            JSON formatted file defining symbols to be exported to the CM3
        v1
            Flag to determine if old style parsing should be performed when
            parsing the memory areas.
        '''
        self.inFile = open(path.realpath(elf.name), "rb")
        self.elf = ELFFile(self.inFile)
        self.output = output
        self.tag = tag
        self.symbols = path.realpath(symbols.name) if symbols else None
        self._memoryIntegrator = DSPMemoryIntegrator(self.elf, output, tag, v1)
        self._symbolIntegrator = DSPSymbolIntegrator(self.elf, output, tag, self.symbols)
        
    def parse(self):
        ''' Parse the memory areas and symbols as required ''' 
        self._memoryIntegrator.parseMemoryAreas()
        if (self.symbols):
            self._symbolIntegrator.parseSymbols()
            
